package qses.apache.commons;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/* Code adapted from the Apache commons library. */
public class InetAddressValidator   {

  private static final int IPV4_MAX_OCTET_VALUE = 255;

  private static final int MAX_UNSIGNED_SHORT = 0xffff;

  private static final int BASE_16 = 16;

  private static final String IPV4_REGEX =
          "^(\\d{1,3})\\.(\\d{1,3})\\.(\\d{1,3})\\.(\\d{1,3})$";

  // Max number of hex groups (separated by :) in an IPV6 address
  private static final int IPV6_MAX_HEX_GROUPS = 8;

  // Max hex digits in each IPv6 group
  private static final int IPV6_MAX_HEX_DIGITS_PER_GROUP = 4;

  /**
   * Singleton instance of this class.
   */
  private static final InetAddressValidator VALIDATOR = new InetAddressValidator();

  /** IPv4 RegexValidator */
  private final Pattern ipv4Validator = Pattern.compile(IPV4_REGEX);

  /**
   * Returns the singleton instance of this validator.
   * @return the singleton instance of this validator
   */
  public static InetAddressValidator getInstance() {
      return VALIDATOR;
  }

  /**
   * Checks if the specified string is a valid IP address.
   * @param inetAddress the string to validate
   * @return true if the string validates as an IP address
   */
  public boolean isValid(String inetAddress) {
      return isValidInet4Address(inetAddress) || isValidInet6Address(inetAddress);
  }

  /**
   * Validates an IPv4 address. Returns true if valid.
   * @param inet4Address the IPv4 address to validate
   * @return true if the argument contains a valid IPv4 address
   */
  public boolean isValidInet4Address(String inet4Address) {
      // verify that address conforms to generic IPv4 format
      Matcher m = ipv4Validator.matcher(inet4Address);

      if (!m.matches()) {
          return false;
      }
      

      // verify that address subgroups are legal
      for (int i = 1 ; i <= 4; i++) {
          String ipSegment = m.group(i);
          if (ipSegment == null || ipSegment.length() == 0) {
              return false;
          }

          int iIpSegment = 0;

          try {
              iIpSegment = Integer.parseInt(ipSegment);
          } catch(NumberFormatException e) {
              return false;
          }

          if (iIpSegment > IPV4_MAX_OCTET_VALUE) {
              return false;
          }

          if (ipSegment.length() > 1 && ipSegment.startsWith("0")) {
              return false;
          }
      }

      return true;
  }

  public static void main(String[] args) {
    InetAddressValidator v = getInstance();
    String[] addresses = { "25.255.255.255",
                           "255.256.2.1"
    };
    for (String a : addresses) {
      System.out.println(a + " -> " + v.isValidInet4Address(a));
    }
  }
  /**
   * Validates an IPv6 address. Returns true if valid.
   * @param inet6Address the IPv6 address to validate
   * @return true if the argument contains a valid IPv6 address
   * 
   * @since 1.4.1
   */
  public boolean isValidInet6Address(String inet6Address) {
      boolean containsCompressedZeroes = inet6Address.contains("::");
      if (containsCompressedZeroes && (inet6Address.indexOf("::") != inet6Address.lastIndexOf("::"))) {
          return false;
      }
      if ((inet6Address.startsWith(":") && !inet6Address.startsWith("::"))
              || (inet6Address.endsWith(":") && !inet6Address.endsWith("::"))) {
          return false;
      }
      String[] octets = inet6Address.split(":");
      if (containsCompressedZeroes) {
          List<String> octetList = new ArrayList<String>(Arrays.asList(octets));
          if (inet6Address.endsWith("::")) {
              // String.split() drops ending empty segments
              octetList.add("");
          } else if (inet6Address.startsWith("::") && !octetList.isEmpty()) {
              octetList.remove(0);
          }
          octets = octetList.toArray(new String[octetList.size()]);
      }
      if (octets.length > IPV6_MAX_HEX_GROUPS) {
          return false;
      }
      int validOctets = 0;
      int emptyOctets = 0; // consecutive empty chunks
      for (int index = 0; index < octets.length; index++) {
          String octet = octets[index];
          if (octet.length() == 0) {
              emptyOctets++;
              if (emptyOctets > 1) {
                  return false;
              }
          } else {
              emptyOctets = 0;
              // Is last chunk an IPv4 address?
              if (index == octets.length - 1 && octet.contains(".")) {
                  if (!isValidInet4Address(octet)) {
                      return false;
                  }
                  validOctets += 2;
                  continue;
              }
              if (octet.length() > IPV6_MAX_HEX_DIGITS_PER_GROUP) {
                  return false;
              }
              int octetInt = 0;
              try {
                  octetInt = Integer.parseInt(octet, BASE_16);
              } catch (NumberFormatException e) {
                  return false;
              }
              if (octetInt < 0 || octetInt > MAX_UNSIGNED_SHORT) {
                  return false;
              }
          }
          validOctets++;
      }
      if (validOctets > IPV6_MAX_HEX_GROUPS || (validOctets < IPV6_MAX_HEX_GROUPS && !containsCompressedZeroes)) {
          return false;
      }
      return true;
  }
}